# Generated by Django 5.1.1 on 2024-09-05 22:20

from django.db import migrations


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('otp_totp', '0003_add_timestamps'),
    ]

    operations = [
        migrations.CreateModel(
            name='TOTPDeviceProxy',
            fields=[],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('otp_totp.totpdevice',),
        ),
    ]
